/** *************************************************************************
 **
 ** Copyright (C) 2013 Heinz van Saanen
 **
 ** This file is part of the function viewer Zhu3D.
 **
 ** This file may be used under the terms of the GNU General Public
 ** License version 3 as published by the Free Software Foundation
 ** and appearing in the file LICENSE.GPL included in the packaging of
 ** this file.
 **
 ** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 ** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 **
****************************************************************************/

#include <cmath>
#include <ctime>
#include "pinterface.h"
#include "property.h"
#include "qadran.h"


/** NOTE: For speed reasons this module assumes, that FPFLOAT = float */


/** Global parsers
 ** 0..2 Functions
 ** 3..5 Parametric system
 ** 6..8 Isosurfaces */
FunctionParser fpGL[PCOUNT];


/** List for cleaned user entries. Used in the thread-safe iso-parts */
QList<QString> isoList;


/** Counter gets 0..MAXUSR-1 during execution
 ** Used in mainwindow, usredit, gldraw for thread control */
int funpCount=0;


/** Parsers for new constants and functions */
static FunctionParser virp;            // A "virgin" parser used for initializing
static FunctionParser conp;            // The parser holding user defined constants
static FunctionParser funp[MAXUSR];    // The parsers holding user defined functions


/** Fetch entries from user item table */
static int fetchNewConstant(const QString &s);
static int fetchNewFunction(const QString &s);
static int fetchItem(const QString &s);


/** Local helpers */
static bool addParserConstant(const std::string &name, FPFLOAT val);
static bool addDefParserFunction(const std::string &name, FPFLOAT (*funPtr)(const FPFLOAT*), int parAmount);
static bool addNewParserFunction(const std::string &name, FunctionParser &toAdd);
static int updateAllParsers(void);
static QADRAN ran;


/** *************************************************************************
 ** Local default function stuff used in every parser
****************************************************************************/
inline static FPFLOAT sig(const FPFLOAT *p) {
	if(p[0]>0.0f) return 1.0f;
	if(p[0]<0.0f) return -1.0f;
	return 0.0f;
}

inline static FPFLOAT frac(const FPFLOAT *p) {
	FPFLOAT tmp=p[0];
	return tmp-=(FPFLOAT)((int)tmp);
}

inline static FPFLOAT log2(const FPFLOAT *p) {
	return log10f(p[0])/log10f(2.0f);
}

inline static FPFLOAT power(const FPFLOAT *p) {
	return powf(p[0], p[1]);
}

inline static FPFLOAT modul(const FPFLOAT *p) {
	return (FPFLOAT)((int)p[0]%(int)p[1]);
}

inline static FPFLOAT gamma(const FPFLOAT *p) {
	return tgammaf(p[0]);
}

inline static FPFLOAT rand(const FPFLOAT *p) {
	return ((float)ran.frand()*2.0f-1.0f)*p[0];
}


/** *************************************************************************
 ** New user defined constants and functions
****************************************************************************/

/** Fetch constant; "name=vals" */
static int fetchNewConstant(const QString &s) {

	const QString name(s.section('=',0,0));
	const QString vals(s.section('=',1,1));
	FPFLOAT vars[1], fval;
	int err;

	/** Never redefine built in constants
	 ** "ERROR 13: Adding new constant failed" */
	if(name=="_pi" || name=="_e" || name=="_c")
		return 13;

	/** Extract value */
	conp.Parse(q2cstrcpy(vals), "");
	if((err=conp.GetParseErrorType()) != FunctionParser::FP_NO_ERROR)
		return err;
	fval=conp.Eval(vars);

	/** "ERROR 13: Adding new constant failed"  */
	if(!addParserConstant(q2cstrcpy(name), fval))
		return 13;

	/** Everything was ok so far */
	isoList.append(s);
	return FunctionParser::FP_NO_ERROR;
}

/** Fetch function; "name(pars,...)=funs" */
static int fetchNewFunction(const QString &s) {

	const QString name(s.section('=',0,0).section('(',0,0));
	const QString pars(s.section(')',0,0).section('(',1,1));
	const QString funs(s.section('=',1,1));
	char cfuns[funs.length()+1];
	int err;

	/** "ERROR 14: Adding new function failed"
	 ** Too much user entries made so far */
	if(funpCount==(MAXUSR-1))
		return 14;

	/** Check syntax for new entry */
	q2cstrcpy(cfuns, funs);
	funp[funpCount].Parse(cfuns, q2cstrcpy(pars));
	if((err=funp[funpCount].GetParseErrorType()) != FunctionParser::FP_NO_ERROR)
		return err;

	/** "ERROR 14: Adding new function failed" */
	if(!addNewParserFunction(q2cstrcpy(name), funp[funpCount]))
		return 14;

	/** Everything was ok so far */
	funpCount++;
	isoList.append(s);
	return FunctionParser::FP_NO_ERROR;
}

/** Returns FunctionParser::FP_NO_ERROR on success, error number else */
int fetchItem(const QString &s) {
	QString tmp;

	/** Check comment or empty string */
	if(s.isEmpty() || s[0]==QChar('#'))
		return FunctionParser::FP_NO_ERROR;

	/** "ERROR 12: Missing equal sign" */
	if(!s.contains(QChar('=')))
		return 12;

	/** Clean usr-string */
	tmp=prepareInput(s);

	/** Check for constant first */
	if(!tmp.contains(")="))
		return fetchNewConstant(tmp);

	/** Else item must be a function */
	return fetchNewFunction(tmp);
}


/** *************************************************************************
 ** Adding stuff
****************************************************************************/

/** Add a constant to all parsers */
static bool addParserConstant(const std::string &name, FPFLOAT val) {
	long i;

	if(!conp.AddConstant(name, val))
		return false;

	for(i=0; i<lastUsrEntry; i++)
		if(!funp[i].AddConstant(name, val))
			return false;

	for(i=0; i<PCOUNT; i++)
		if(!fpGL[i].AddConstant(name, val))
			return false;

	return true;
}

/** Add a default function to all parsers */
static bool addDefParserFunction(const std::string &name, FPFLOAT (*funPtr)(const FPFLOAT*), int parAmount) {
	long i;

	for(i=0; i<PCOUNT; i++)
		if(!fpGL[i].AddFunction(name, funPtr, (unsigned)parAmount))
			return false;

	for(i=0; i<lastUsrEntry; i++)
		if(!funp[i].AddFunction(name, funPtr, (unsigned)parAmount))
			return false;

	return true;
}

/** Add a new function to all parsers */
static bool addNewParserFunction(const std::string &name, FunctionParser &toAdd) {
	long i;

	if(!conp.AddFunction(name, toAdd))
		return false;

	for(i=funpCount+1; i<lastUsrEntry; i++)
		if(!funp[i].AddFunction(name, toAdd))
			return false;

	for(i=0; i<PCOUNT; i++)
		if(!fpGL[i].AddFunction(name, toAdd))
			return false;

	return true;
}


/** *************************************************************************
 ** Central parser stuff
****************************************************************************/

/** Called when loading a new file or when making user defined entries */
int initAllParsers(void) {
	int i, err;

	/** Care for a reasonable starting point with "virgin" parsers */
	for(i=0; i<PCOUNT; i++) fpGL[i]=virp;
	for(i=0; i<MAXUSR; i++) funp[i]=virp;
	conp=virp;
	funpCount=0;
	isoList.clear();

	/** Add all default constants */
	if(!addParserConstant("_pi",(FPFLOAT)3.1415926535897932384)) return 13;
	if(!addParserConstant("_e", (FPFLOAT)2.7182818284590452354)) return 13;
	if(!addParserConstant("_c", (FPFLOAT)299792458.0)) return 13;

	/** Add all default functions */
	if(!addDefParserFunction("sig",   sig,    1)) return 11;
	if(!addDefParserFunction("rnd",   rand,   1)) return 11;
	if(!addDefParserFunction("frac",  frac,   1)) return 11;
	if(!addDefParserFunction("log2",  log2,   1)) return 11;
	if(!addDefParserFunction("pow",   power,  2)) return 11;
	if(!addDefParserFunction("mod",   modul,  2)) return 11;
	if(!addDefParserFunction("gamma", gamma,  1)) return 11;

	/** Fetch user items */
	for(i=0; i<lastUsrEntry; i++)
		if((err=fetchItem(usr[i])) != FunctionParser::FP_NO_ERROR)
			return err;

	/** Update at last */
	if((err=updateAllParsers()) != FunctionParser::FP_NO_ERROR)
		return err;

	/** Everything was ok */
	return FunctionParser::FP_NO_ERROR;
}


#define FPOPTIMIZER
/** *************************************************************************
 ** Needed for opening new files or entering new functions/constants and is
 ** evaluated only once there. Time needed is some ms even for complicated
 ** calculations. So not fiddling arround and update all together
****************************************************************************/
static int updateAllParsers(void) {
	int i, err;

	/** Functions */
	for(i=0; i<3; i++) {
		fpGL[i].Parse(fun[i].str, "x,y");
		if((err=fpGL[i].GetParseErrorType()) != FunctionParser::FP_NO_ERROR) {
			strcpy(fun[i].str, q2cstrcpy(bakfun[i]));
			return err;
		}
		#ifdef FPOPTIMIZER
		else
			fpGL[i].Optimize();
		#endif
	}

	/** Parameter */
	for(i=3; i<6; i++) {
		fpGL[i].Parse(par.str[i%3], "s,t");
		if((err=fpGL[i].GetParseErrorType()) != FunctionParser::FP_NO_ERROR) {
			strcpy(par.str[i%3], q2cstrcpy(bakpar[i%3]));
			return err;
		}
		#ifdef FPOPTIMIZER
		else
			fpGL[i].Optimize();
		#endif
	}

	/** Isosurfaces */
	for(i=6; i<9; i++) {
		fpGL[i].Parse(iso[i%3].str, "x,y,z");
		if((err=fpGL[i].GetParseErrorType()) != FunctionParser::FP_NO_ERROR) {
			strcpy(iso[i%3].str, q2cstrcpy(bakiso[i%3]));
			return err;
		}
		#ifdef FPOPTIMIZER
		else
			fpGL[i].Optimize();
		#endif
	}

	return FunctionParser::FP_NO_ERROR;
}


/** *************************************************************************
 ** Fetch and retranslate error messages
****************************************************************************/
const char *fpError::eString[] = {
	"ERROR 0: Syntax error",
	"ERROR 1: Mismatched parenthesis",
	"ERROR 2: Missing closing parenthesis",
	"ERROR 3: Empty parentheses",
	"ERROR 4: Operator expected",
	"ERROR 5: Out of memory",
	"ERROR 6: Unexpected error",
	"ERROR 7: Unknown parameter",
	"ERROR 8: Illegal number of function parameters",
	"ERROR 9: Premature end of string",
	"ERROR 10: Expecting parenthesis after function",
	"ERROR 11: Parser initialization failed",
	"ERROR 12: Missing equal sign",
	"ERROR 13: Adding new constant failed",
	"ERROR 14: Adding new function failed",
	"ERROR 15: Parser update failed"
};

QString fpError::eQString[] = {
	QObject::tr("ERROR 0: Syntax error"),
	QObject::tr("ERROR 1: Mismatched parenthesis"),
	QObject::tr("ERROR 2: Missing closing parenthesis"),
	QObject::tr("ERROR 3: Empty parentheses"),
	QObject::tr("ERROR 4: Operator expected"),
	QObject::tr("ERROR 5: Out of memory"),
	QObject::tr("ERROR 6: Unexpected error"),
	QObject::tr("ERROR 7: Unknown parameter"),
	QObject::tr("ERROR 8: Illegal number of function parameters"),
	QObject::tr("ERROR 9: Premature end of string"),
	QObject::tr("ERROR 10: Expecting parenthesis after function"),
	QObject::tr("ERROR 11: Parser initialization failed"),
	QObject::tr("ERROR 12: Missing equal sign"),
	QObject::tr("ERROR 13: Adding new constant failed"),
	QObject::tr("ERROR 14: Adding new function failed"),
	QObject::tr("ERROR 15: Parser update failed")
};

void fpError::initParserQt() {
	for(long i=0; i<=15; i++)
		eQString[i]=QObject::tr(eString[i]);
}
